@@ -51,104 +51,31 @@ namespace libsemigroups {
   ////////////////////////////////////////////////////////////////////////
 
   TEMPLATE
-  FROIDURE_PIN::FroidurePin(std::vector<element_type> const* gens)
+  FROIDURE_PIN::FroidurePin()
       : detail::BruidhinnTraits<TElementType>(),
         FroidurePinBase(),
-        _degree(UNDEFINED),
-        _duplicate_gens(),
         _elements(),
-        _enumerate_order(),
-        _final(),
-        _first(),
-        _found_one(false),
         _gens(),
         _id(),
-        _idempotents(),
-        _idempotents_found(false),
-        _is_idempotent(),
-        _left(gens->size()),
-        _length(),
-        _lenindex(),
-        _letter_to_pos(),
         _map(),
         _mtx(),
-        _nr(0),
-        _nrgens(gens->size()),
-        _nr_rules(0),
-        _pos(0),
-        _pos_one(0),
-        _prefix(),
-        _reduced(gens->size()),
-        _relation_gen(0),
-        _relation_pos(UNDEFINED),
-        _right(gens->size()),
         _sorted(),
-        _suffix(),
-        _tmp_product(),
-        _wordlen(0) {  // (length of the current word) - 1
-    if (_nrgens == 0) {
-      LIBSEMIGROUPS_EXCEPTION("no generators given");
-    }
+        _state(nullptr),
+        _tmp_product() {  // (length of the current word) - 1
 #ifdef LIBSEMIGROUPS_VERBOSE
     _nr_products = 0;
 #endif
     _right.set_default_value(UNDEFINED);
-    // FIXME(later) inclusion of the next line makes test FroidurePin of BMats
-    // 01 extremely slow (~50ms to ~10s!!!!)
-    // reserve(_nrgens);
-    _degree = Degree()((*gens)[0]);
+  }
 
-    for (size_t i = 0; i < _nrgens; ++i) {
-      size_t degree = Degree()((*gens)[i]);
-      if (degree != _degree) {
-        LIBSEMIGROUPS_EXCEPTION(
-            "generator %d has degree %d but should have degree %d",
-            i,
-            degree,
-            _degree);
-      }
-    }
-    for (const_reference x : *gens) {
-      _gens.push_back(this->internal_copy(this->to_internal_const(x)));
-    }
-
-    _tmp_product = this->to_internal(One()(this->to_external_const(_gens[0])));
-    _id          = this->to_internal(One()(this->to_external_const(_gens[0])));
-    _lenindex.push_back(0);
-
-    // #ifdef LIBSEMIGROUPS_DENSEHASHMAP
-    //     _map.set_empty_key(this->empty_key(_id));
-    // #endif
-
-    // add the generators
-    for (letter_type i = 0; i < _nrgens; i++) {
-      auto it = _map.find(_gens[i]);
-      if (it != _map.end()) {  // duplicate generator
-        _letter_to_pos.push_back(it->second);
-        _nr_rules++;
-        _duplicate_gens.emplace_back(i, _first[it->second]);
-        // i.e. _gens[i] = _gens[_first[it->second]]
-        // _first maps from element_index_type -> letter_type :)
-      } else {
-        is_one(_gens[i], _nr);
-        _elements.push_back(_gens[i]);
-        // Note that every non-duplicate generator is *really* stored in
-        // _elements, and so must be deleted from _elements but not _gens.
-        _first.push_back(i);
-        _final.push_back(i);
-        _enumerate_order.push_back(_nr);
-        _letter_to_pos.push_back(_nr);
-        _length.push_back(1);
-        _map.emplace(_elements.back(), _nr);
-        _prefix.push_back(UNDEFINED);
-        // TODO(later) _prefix.push_back(_nr) and get rid of _letter_to_pos, and
-        // the extra clause in the run member function!
-        _suffix.push_back(UNDEFINED);
-        _nr++;
-      }
-    }
-    expand(_nr);
-    _lenindex.push_back(_enumerate_order.size());
+  TEMPLATE
+  FROIDURE_PIN::FroidurePin(std::vector<element_type> const* gens)
+      : FroidurePin() {
+#ifdef LIBSEMIGROUPS_VERBOSE
+    _nr_products = 0;
+#endif
+    validate_element_collection(gens->cbegin(), gens->cend());
+    add_generators_before_start(gens->cbegin(), gens->cend());
   }
 
   TEMPLATE
@@ -163,63 +90,32 @@ namespace libsemigroups {
   FROIDURE_PIN::FroidurePin(FroidurePin const& S)
       : detail::BruidhinnTraits<TElementType>(),
         FroidurePinBase(S),
-        _degree(S._degree),
-        _duplicate_gens(S._duplicate_gens),
         _elements(),
-        _enumerate_order(S._enumerate_order),
-        _final(S._final),
-        _first(S._first),
-        _found_one(S._found_one),
         _gens(),
-        _id(this->internal_copy(S._id)),
+        _id(),
         _idempotents(S._idempotents),
-        _idempotents_found(S._idempotents_found),
-        _is_idempotent(S._is_idempotent),
-        _left(S._left),
-        _length(S._length),
-        _lenindex(S._lenindex),
-        _letter_to_pos(S._letter_to_pos),
-        _nr(S._nr),
-        _nrgens(S._nrgens),
-        _nr_rules(S._nr_rules),
-        _pos(S._pos),
-        _pos_one(S._pos_one),
-        _prefix(S._prefix),
-        _reduced(S._reduced),
-        _relation_gen(S._relation_gen),
-        _relation_pos(S._relation_pos),
-        _right(S._right),
         _sorted(),  // TODO(later) S this if set
-        _suffix(S._suffix),
-        _wordlen(S._wordlen) {
-#ifdef LIBSEMIGROUPS_VERBOSE
-    _nr_products = 0;
-#endif
+        _state(S._state) {
     _elements.reserve(_nr);
 
-    // #ifdef LIBSEMIGROUPS_DENSEHASHMAP
-    //     _map.set_empty_key(this->empty_key(_id));
-    //     _map.resize(_nr);
-    // #else
-    //     // _map.reserve(_nr); // FIXME(later) uncommenting this line causes
-    //     huge
-    //     // slow down in FroidurePin 000
-    // #endif
-    _tmp_product = this->internal_copy(S._id);
-
     element_index_type i = 0;
     for (internal_const_reference x : S._elements) {
       auto y = this->internal_copy(x);
       _elements.push_back(y);
       _map.emplace(y, i++);
     }
-    copy_gens();
+    if (S.nr_generators() != 0) {
+      copy_generators_from_elements(S.nr_generators());
+      init_degree(this->to_external_const(_gens[0]));
+    }
   }
 
   TEMPLATE
   FROIDURE_PIN::~FroidurePin() {
-    this->internal_free(_tmp_product);
-    this->internal_free(_id);
+    if (nr_generators() > 0) {
+      this->internal_free(_tmp_product);
+      this->internal_free(_id);
+    }
 
     // delete those generators not in _elements, i.e. the duplicate ones
     for (auto& x : _duplicate_gens) {
@@ -256,81 +152,32 @@ namespace libsemigroups {
   TEMPLATE
   FROIDURE_PIN::FroidurePin(FroidurePin const&               S,
                             std::vector<element_type> const* coll)
-      : _degree(S._degree),  // copy for comparison in add_generators
-        _duplicate_gens(S._duplicate_gens),
-        _elements(),
-        _found_one(S._found_one),  // copy in case degree doesn't change in
-                                   // add_generators
-        _gens(),
-        _idempotents(S._idempotents),
-        _idempotents_found(S._idempotents_found),
-        _is_idempotent(S._is_idempotent),
-        _left(S._left),
-        _letter_to_pos(S._letter_to_pos),
-        _nr(S._nr),
-        _nrgens(S._nrgens),
-        _nr_rules(0),
-        _pos(S._pos),
-        _pos_one(S._pos_one),  // copy in case degree doesn't change in
-                               // add_generators
-        _reduced(S._reduced),
-        _relation_gen(0),
-        _relation_pos(UNDEFINED),
-        _right(S._right),
-        _sorted(),
-        _wordlen(0) {
+      : FroidurePin() {
+    _idempotents = S._idempotents;
+    _state       = S._state;
+    LIBSEMIGROUPS_ASSERT(S._lenindex.size() > 1);
     LIBSEMIGROUPS_ASSERT(!coll->empty());
     LIBSEMIGROUPS_ASSERT(Degree()(coll->at(0)) >= S.degree());
+    partial_copy(S);
 
 #ifdef LIBSEMIGROUPS_DEBUG
     for (const_reference x : *coll) {
       LIBSEMIGROUPS_ASSERT(Degree()(x) == Degree()((*coll)[0]));
     }
-#endif
-#ifdef LIBSEMIGROUPS_VERBOSE
-    _nr_products = 0;
 #endif
     _elements.reserve(S._nr);
 
-    // the following are required for assignment to specific positions in
-    // add_generators
-    _final.resize(S._nr, 0);
-    _first.resize(S._nr, 0);
-    _length.resize(S._nr, 0);
-    _prefix.resize(S._nr, 0);
-    _suffix.resize(S._nr, 0);
-
     size_t deg_plus = Degree()(coll->at(0)) - S.degree();
-
     if (deg_plus != 0) {
       _degree += deg_plus;
       _found_one = false;
       _pos_one   = 0;
     }
 
-    _lenindex.push_back(0);
-    _lenindex.push_back(S._lenindex[1]);
-    _enumerate_order.reserve(S._nr);
-
-    // add the distinct old generators to new _enumerate_order
-    for (enumerate_index_type i = 0; i < S._lenindex[1]; i++) {
-      _enumerate_order.push_back(S._enumerate_order[i]);
-      _final[_enumerate_order[i]]  = S._final[S._enumerate_order[i]];
-      _first[_enumerate_order[i]]  = S._first[S._enumerate_order[i]];
-      _prefix[_enumerate_order[i]] = UNDEFINED;
-      _suffix[_enumerate_order[i]] = UNDEFINED;
-      _length[_enumerate_order[i]] = 1;
-    }
-
     _id          = One()(this->to_internal(coll->at(0)));
     _tmp_product = this->internal_copy(_id);
 
-    // #ifdef LIBSEMIGROUPS_DENSEHASHMAP
-    //     _map.set_empty_key(this->empty_key(_id));
-    //     _map.resize(S._nr);
-    // #else
     _map.reserve(S._nr);
-    // #endif
 
     element_index_type i = 0;
     for (internal_const_reference x : S._elements) {
@@ -340,7 +187,8 @@ namespace libsemigroups {
       _map.emplace(y, i);
       is_one(y, i++);
     }
-    copy_gens();  // copy the old generators
+    // copy the old generators
+    copy_generators_from_elements(S.nr_generators());
     // Now this is ready to have add_generators or closure called on it
   }
 
@@ -348,21 +196,6 @@ namespace libsemigroups {
   // FroidurePin - member functions - public
   ////////////////////////////////////////////////////////////////////////
 
-  ELEMENT_INDEX_TYPE FROIDURE_PIN::word_to_pos(word_type const& w) const {
-    // w is a word in the generators (i.e. a vector of letter_type's)
-    if (w.size() == 0) {
-      LIBSEMIGROUPS_EXCEPTION("the given word has length 0");
-    }
-    for (auto x : w) {
-      validate_letter_index(x);
-    }
-    element_index_type out = _letter_to_pos[w[0]];
-    for (auto it = w.cbegin() + 1; it < w.cend() && out != UNDEFINED; ++it) {
-      out = _right.get(out, _letter_to_pos[*it]);
-    }
-    return out;
-  }
-
   ELEMENT_TYPE FROIDURE_PIN::word_to_element(word_type const& w) const {
     element_index_type pos = word_to_pos(w);
     if (pos != UNDEFINED) {
@@ -375,15 +208,19 @@ namespace libsemigroups {
     LIBSEMIGROUPS_ASSERT(w[0] < nr_generators() && w[1] < nr_generators());
     element_type prod
         = this->external_copy(this->to_external_const(_tmp_product));
-    Product()(prod,
-              this->to_external_const(_gens[w[0]]),
-              this->to_external_const(_gens[w[1]]));
+
+    auto* state_ptr = _state.get();
+    InternalProduct()(prod,
+                      this->to_external_const(_gens[w[0]]),
+                      this->to_external_const(_gens[w[1]]),
+                      state_ptr);
     for (auto it = w.begin() + 2; it < w.end(); ++it) {
       LIBSEMIGROUPS_ASSERT(*it < nr_generators());
       Swap()(this->to_external(_tmp_product), prod);
-      Product()(prod,
-                this->to_external_const(_tmp_product),
-                this->to_external_const(_gens[*it]));
+      InternalProduct()(prod,
+                        this->to_external_const(_tmp_product),
+                        this->to_external_const(_gens[*it]),
+                        state_ptr);
     }
     return prod;
   }
@@ -405,14 +242,6 @@ namespace libsemigroups {
     }
   }
 
-  SIZE_T FROIDURE_PIN::current_max_word_length() const noexcept {
-    return _length[_enumerate_order.back()];
-  }
-
-  SIZE_T FROIDURE_PIN::degree() const noexcept {
-    return _degree;
-  }
-
   SIZE_T FROIDURE_PIN::nr_generators() const noexcept {
     return _gens.size();
   }
@@ -431,98 +260,29 @@ namespace libsemigroups {
     return (it == _map.end() ? UNDEFINED : it->second);
   }
 
-  SIZE_T FROIDURE_PIN::current_size() const noexcept {
-    return _elements.size();
-  }
-
-  SIZE_T FROIDURE_PIN::current_nr_rules() const noexcept {
-    return _nr_rules;
-  }
-
-  ELEMENT_INDEX_TYPE
-  FROIDURE_PIN::prefix(element_index_type pos) const {
-    validate_element_index(pos);
-    return _prefix[pos];
-  }
-
-  ELEMENT_INDEX_TYPE
-  FROIDURE_PIN::suffix(element_index_type pos) const {
-    validate_element_index(pos);
-    return _suffix[pos];
-  }
-
-  LETTER_TYPE
-  FROIDURE_PIN::first_letter(element_index_type pos) const {
-    validate_element_index(pos);
-    return _first[pos];
-  }
-
-  LETTER_TYPE
-  FROIDURE_PIN::final_letter(element_index_type pos) const {
-    validate_element_index(pos);
-    return _final[pos];
-  }
-
-  SIZE_T FROIDURE_PIN::length_const(element_index_type pos) const {
-    validate_element_index(pos);
-    return _length[pos];
-  }
-
-  SIZE_T FROIDURE_PIN::length_non_const(element_index_type pos) {
-    if (pos >= _nr) {
-      run();
-    }
-    return length_const(pos);
-  }
-
-  ELEMENT_INDEX_TYPE
-  FROIDURE_PIN::product_by_reduction(element_index_type i,
-                                     element_index_type j) const {
-    validate_element_index(i);
-    validate_element_index(j);
-
-    if (length_const(i) <= length_const(j)) {
-      while (i != UNDEFINED) {
-        j = _left.get(j, _final[i]);
-        i = _prefix[i];
-      }
-      return j;
-    } else {
-      while (j != UNDEFINED) {
-        i = _right.get(i, _first[j]);
-        j = _suffix[j];
-      }
-      return i;
-    }
-  }
-
   ELEMENT_INDEX_TYPE
   FROIDURE_PIN::fast_product(element_index_type i, element_index_type j) const {
     validate_element_index(i);
     validate_element_index(j);
-    if (length_const(i)
-            < 2 * Complexity()(this->to_external_const(_tmp_product))
-        || length_const(j)
-               < 2 * Complexity()(this->to_external_const(_tmp_product))) {
+    auto const n = 2 * Complexity()(this->to_external_const(_tmp_product));
+    if (length_const(i) < n || length_const(j) < n) {
       return product_by_reduction(i, j);
     } else {
-      Product()(this->to_external(_tmp_product),
-                this->to_external_const(_elements[i]),
-                this->to_external_const(_elements[j]));
+      InternalProduct()(this->to_external(_tmp_product),
+                        this->to_external_const(_elements[i]),
+                        this->to_external_const(_elements[j]),
+                        _state.get());
       return _map.find(_tmp_product)->second;
     }
   }
 
-  ELEMENT_INDEX_TYPE FROIDURE_PIN::letter_to_pos(letter_type i) const {
-    validate_letter_index(i);
-    return _letter_to_pos[i];
-  }
-
+  // TODO(later) put this in FroidurePinBase??
   SIZE_T FROIDURE_PIN::nr_idempotents() {
     init_idempotents();
     return _idempotents.size();
   }
 
+  // TODO(later) put this in FroidurePinBase??
   BOOL FROIDURE_PIN::is_idempotent(element_index_type pos) {
     init_idempotents();
     // only validate pos after init_idempotents, because we don't know if it's
@@ -531,11 +291,6 @@ namespace libsemigroups {
     return _is_idempotent[pos];
   }
 
-  SIZE_T FROIDURE_PIN::nr_rules() {
-    run();
-    return _nr_rules;
-  }
-
   VOID FROIDURE_PIN::reserve(size_t n) {
     // Since the FroidurePin we are enumerating is bounded in size by the
     // maximum value of an element_index_t, we cast the argument here to this
@@ -547,24 +302,13 @@ namespace libsemigroups {
     _enumerate_order.reserve(nn);
     _left.reserve(nn);
     _length.reserve(nn);
-
-    // #ifdef LIBSEMIGROUPS_DENSEHASHMAP
-    //    _map.resize(nn);
-    // #else
     _map.reserve(nn);
-    // #endif
-
     _prefix.reserve(nn);
     _reduced.reserve(nn);
     _right.reserve(nn);
     _suffix.reserve(nn);
   }
 
-  SIZE_T FROIDURE_PIN::size() {
-    run();
-    return _elements.size();
-  }
-
   BOOL FROIDURE_PIN::contains(const_reference x) {
     return (position(x) != UNDEFINED);
   }
@@ -616,52 +360,6 @@ namespace libsemigroups {
     return this->to_external_const(_sorted.at(pos).first);
   }
 
-  ELEMENT_INDEX_TYPE FROIDURE_PIN::right(element_index_type i, letter_type j) {
-    validate_letter_index(j);
-    run();
-    validate_element_index(i);
-    return _right.get(i, j);
-  }
-
-  CAYLEY_GRAPH_TYPE const& FROIDURE_PIN::right_cayley_graph() {
-    run();
-    _right.shrink_rows_to(size());
-    return _right;
-  }
-
-  ELEMENT_INDEX_TYPE FROIDURE_PIN::left(element_index_type i, letter_type j) {
-    validate_letter_index(j);
-    run();
-    validate_element_index(i);
-    return _left.get(i, j);
-  }
-
-  CAYLEY_GRAPH_TYPE const& FROIDURE_PIN::left_cayley_graph() {
-    run();
-    _left.shrink_rows_to(size());
-    return _left;
-  }
-
-  VOID FROIDURE_PIN::minimal_factorisation(word_type&         word,
-                                           element_index_type pos) {
-    if (pos >= _nr && !finished()) {
-      enumerate(pos + 1);
-    }
-    validate_element_index(pos);
-    word.clear();
-    while (pos != UNDEFINED) {
-      word.push_back(_first[pos]);
-      pos = _suffix[pos];
-    }
-  }
-
-  WORD_TYPE
-  FROIDURE_PIN::minimal_factorisation(element_index_type pos) {
-    word_type word;
-    minimal_factorisation(word, pos);
-    return word;
-  }
-
   WORD_TYPE FROIDURE_PIN::minimal_factorisation(const_reference x) {
     element_index_type pos = this->position(x);
     if (pos == UNDEFINED) {
@@ -671,83 +369,10 @@ namespace libsemigroups {
     return minimal_factorisation(pos);
   }
 
-  VOID FROIDURE_PIN::factorisation(word_type& word, element_index_type pos) {
-    minimal_factorisation(word, pos);
-  }
-
-  WORD_TYPE FROIDURE_PIN::factorisation(element_index_type pos) {
-    return minimal_factorisation(pos);
-  }
-
   WORD_TYPE FROIDURE_PIN::factorisation(const_reference x) {
     return minimal_factorisation(x);
   }
 
-  VOID FROIDURE_PIN::reset_next_relation() noexcept {
-    _relation_pos = UNDEFINED;
-    _relation_gen = 0;
-  }
-
-  VOID FROIDURE_PIN::next_relation(word_type& relation) {
-    if (!finished()) {
-      run();
-    }
-
-    relation.clear();
-
-    if (_relation_pos == _nr) {  // no more relations
-      return;
-    }
-
-    if (_relation_pos != UNDEFINED) {
-      while (_relation_pos < _nr) {
-        while (_relation_gen < _nrgens) {
-          if (!_reduced.get(_enumerate_order[_relation_pos], _relation_gen)
-              && (_relation_pos < _lenindex[1]
-                  || _reduced.get(_suffix[_enumerate_order[_relation_pos]],
-                                  _relation_gen))) {
-            relation.push_back(_enumerate_order[_relation_pos]);
-            relation.push_back(_relation_gen);
-            relation.push_back(
-                _right.get(_enumerate_order[_relation_pos], _relation_gen));
-            break;
-          }
-          _relation_gen++;
-        }
-        if (_relation_gen == _nrgens) {  // then relation is empty
-          _relation_gen = 0;
-          _relation_pos++;
-        } else {
-          break;
-        }
-      }
-      _relation_gen++;
-    } else {
-      // duplicate generators
-      if (_relation_gen < _duplicate_gens.size()) {
-        relation.push_back(_duplicate_gens[_relation_gen].first);
-        relation.push_back(_duplicate_gens[_relation_gen].second);
-        _relation_gen++;
-      } else {
-        _relation_gen = 0;
-        _relation_pos++;
-        next_relation(relation);
-      }
-    }
-  }
-
-  VOID FROIDURE_PIN::enumerate(size_t limit) {
-    if (finished() || limit <= current_size()) {
-      return;
-    } else if (LIMIT_MAX - batch_size() > current_size()) {
-      limit = std::max(limit, current_size() + batch_size());
-    } else {  // batch_size() is very big for some reason
-      limit = batch_size();
-    }
-    REPORT_DEFAULT("limit = %llu (%s)\n", limit, __func__);
-    run_until([this, &limit]() -> bool { return current_size() >= limit; });
-  }
-
   VOID FROIDURE_PIN::run_impl() {
     std::lock_guard<std::mutex> lg(_mtx);
     if (_pos >= _nr) {
@@ -757,16 +382,21 @@ namespace libsemigroups {
     detail::Timer timer;
     size_t        tid = THREAD_ID_MANAGER.tid(std::this_thread::get_id());
 
+    LIBSEMIGROUPS_ASSERT(_lenindex.size() > 1);
+
+    auto ptr = _state.get();
+
     // product the generators by every generator
     if (_pos < _lenindex[1]) {
       size_type nr_shorter_elements = _nr;
       while (_pos < _lenindex[1]) {
         element_index_type i = _enumerate_order[_pos];
-        for (letter_type j = 0; j != _nrgens; ++j) {
-          Product()(this->to_external(_tmp_product),
-                    this->to_external_const(_elements[i]),
-                    this->to_external_const(_gens[j]),
-                    tid);
+        for (letter_type j = 0; j != nr_generators(); ++j) {
+          InternalProduct()(this->to_external(_tmp_product),
+                            this->to_external_const(_elements[i]),
+                            this->to_external_const(_gens[j]),
+                            ptr,
+                            tid);
 
 #ifdef LIBSEMIGROUPS_VERBOSE
           _nr_products++;
@@ -795,7 +425,7 @@ namespace libsemigroups {
       }
       for (enumerate_index_type i = 0; i != _pos; ++i) {
         letter_type b = _final[_enumerate_order[i]];
-        for (letter_type j = 0; j != _nrgens; ++j) {
+        for (letter_type j = 0; j != nr_generators(); ++j) {
           _left.set(_enumerate_order[i], j, _right.get(_letter_to_pos[j], b));
         }
       }
@@ -811,7 +441,7 @@ namespace libsemigroups {
         element_index_type i = _enumerate_order[_pos];
         letter_type        b = _first[i];
         element_index_type s = _suffix[i];
-        for (letter_type j = 0; j != _nrgens; ++j) {
+        for (letter_type j = 0; j != nr_generators(); ++j) {
           if (!_reduced.get(s, j)) {
             element_index_type r = _right.get(s, j);
             if (_found_one && r == _pos_one) {
@@ -822,10 +452,11 @@ namespace libsemigroups {
               _right.set(i, j, _right.get(_letter_to_pos[b], _final[r]));
             }
           } else {
-            Product()(this->to_external(_tmp_product),
-                      this->to_external_const(_elements[i]),
-                      this->to_external_const(_gens[j]),
-                      tid);
+            InternalProduct()(this->to_external(_tmp_product),
+                              this->to_external_const(_elements[i]),
+                              this->to_external_const(_gens[j]),
+                              ptr,
+                              tid);
 #ifdef LIBSEMIGROUPS_VERBOSE
             _nr_products++;
 #endif
@@ -858,7 +489,7 @@ namespace libsemigroups {
         for (enumerate_index_type i = _lenindex[_wordlen]; i != _pos; ++i) {
           element_index_type p = _prefix[_enumerate_order[i]];
           letter_type        b = _final[_enumerate_order[i]];
-          for (letter_type j = 0; j != _nrgens; ++j) {
+          for (letter_type j = 0; j != nr_generators(); ++j) {
             _left.set(_enumerate_order[i], j, _right.get(_left.get(p, j), b));
           }
         }
@@ -881,78 +512,85 @@ namespace libsemigroups {
     return !running() && _pos >= _nr;
   }
 
-  VOID FROIDURE_PIN::add_generator(element_type const& x) {
-    add_generators({x});
+  VOID FROIDURE_PIN::validate_element(const_reference x) const {
+    size_t const n = Degree()(x);
+    if (degree() != UNDEFINED && n != degree()) {
+      LIBSEMIGROUPS_EXCEPTION(
+          "element has degree %d but should have degree %d", n, degree());
+    }
   }
 
   TEMPLATE
-  template <typename TCollection>
-  void FROIDURE_PIN::add_generators(TCollection const& coll) {
-    static_assert(!std::is_pointer<TCollection>::value,
-                  "TCollection should not be a pointer");
-    if (immutable()) {
-      LIBSEMIGROUPS_EXCEPTION("cannot add generators, the FroidurePin instance "
-                              "has been set to immutable");
-    } else if (coll.size() == 0) {
-      return;
-    }
-    for (auto it = coll.begin(); it < coll.end(); ++it) {
-      element_index_type degree = Degree()(*it);
-      if (degree != _degree) {
-        LIBSEMIGROUPS_EXCEPTION(
-            "new generator %llu has degree %llu but should have degree %llu",
-            it - coll.begin(),
-            degree,
-            _degree);
+  template <typename T>
+  void FROIDURE_PIN::validate_element_collection(T const& first,
+                                                 T const& last) const {
+    if (degree() == UNDEFINED && std::distance(first, last) != 0) {
+      auto const n = Degree()(*first);
+      for (auto it = first + 1; it < last; ++it) {
+        auto const m = Degree()(*it);
+        if (m != n) {
+          LIBSEMIGROUPS_EXCEPTION(
+              "element has degree %d but should have degree %d", n, m);
+        }
+      }
+    } else {
+      for (auto it = first; it < last; ++it) {
+        validate_element(*it);
       }
     }
-    detail::Timer timer;
-    size_t        tid = THREAD_ID_MANAGER.tid(std::this_thread::get_id());
+  }
 
-    // get some parameters from the old semigroup
-    letter_type old_nrgens  = _nrgens;
-    size_type   old_nr      = _nr;
-    size_type   nr_old_left = _pos;
+  VOID FROIDURE_PIN::init_degree(const_reference x) {
+    if (_degree == UNDEFINED) {
+      _degree      = Degree()(x);
+      _id          = this->to_internal(One()(x));
+      _tmp_product = this->to_internal(One()(x));
+    }
+  }
 
-    // erase the old index
-    _enumerate_order.erase(_enumerate_order.begin() + _lenindex[1],
-                           _enumerate_order.end());
-
-    // old_new[i] indicates if we have seen _elements.at(i) yet in new.
-    std::vector<bool> old_new;
-    old_new.clear();
-    old_new.resize(old_nr, false);
-    for (letter_type i = 0; i < _letter_to_pos.size(); i++) {
-      old_new[_letter_to_pos[i]] = true;
+  TEMPLATE
+  template <typename T>
+  void FROIDURE_PIN::add_generators_before_start(T const& first,
+                                                 T const& last) {
+    size_t const m = std::distance(first, last);
+    if (m != 0) {
+      init_degree(*first);
     }
 
-    // add the new generators to new _gens, _elements, and _enumerate_order
-    for (const_reference x : coll) {
-      auto it = _map.find(this->to_internal_const(x));
-      if (it == _map.end()) {  // new generator
-        _gens.push_back(this->internal_copy(this->to_internal_const(x)));
+    size_t nr_new_elements = 0;
+
+    for (auto it_coll = first; it_coll < last; ++it_coll) {
+      auto it = _map.find(this->to_internal_const(*it_coll));
+      if (it == _map.end()) {
+        // new generator
+        nr_new_elements++;
+        _gens.push_back(this->internal_copy(this->to_internal_const(*it_coll)));
+        letter_type const n = _gens.size() - 1;
+
+        is_one(_gens.back(), _nr);
         _elements.push_back(_gens.back());
-        _map.emplace(_gens.back(), _nr);
-
-        _first.push_back(_gens.size() - 1);
-        _final.push_back(_gens.size() - 1);
-
-        _letter_to_pos.push_back(_nr);
         _enumerate_order.push_back(_nr);
-
-        is_one(this->to_internal_const(x), _nr);
+        // Note that every non-duplicate generator is *really* stored in
+        // _elements, and so must be deleted from _elements but not _gens.
+        _first.push_back(n);
+        _final.push_back(n);
+        _letter_to_pos.push_back(_nr);
+        _length.push_back(1);
+        _map.emplace(_elements.back(), _nr);
         _prefix.push_back(UNDEFINED);
         _suffix.push_back(UNDEFINED);
-        _length.push_back(1);
         _nr++;
-      } else if (_letter_to_pos[_first[it->second]] == it->second) {
-        _gens.push_back(this->internal_copy(this->to_internal_const(x)));
-        // x is one of the existing generators
-        _duplicate_gens.push_back(
-            std::make_pair(_gens.size() - 1, _first[it->second]));
-        // _gens[_gens.size() - 1] = _gens[_first[it->second])]
-        // since _first maps element_index_type -> letter_type
+        // TODO(later) _prefix.push_back(_nr) and get rid of _letter_to_pos, and
+        // the extra clause in the run member function!
+      } else if (!started()
+                 || _letter_to_pos[_first[it->second]] == it->second) {
+        // duplicate generator
+        // i.e. _gens[i] = _gens[_first[it->second]]
+        // _first maps from element_index_type -> letter_type :)
         _letter_to_pos.push_back(it->second);
+        _nr_rules++;
+        _duplicate_gens.emplace_back(_gens.size(), _first[it->second]);
+        _gens.push_back(this->internal_copy(this->to_internal_const(*it_coll)));
       } else {
         // x is an old element that will now be a generator
         _gens.push_back(_elements[it->second]);
@@ -964,33 +602,55 @@ namespace libsemigroups {
         _prefix[it->second] = UNDEFINED;
         _suffix[it->second] = UNDEFINED;
         _length[it->second] = UNDEFINED;
-
-        old_new[it->second] = true;
       }
     }
+    expand(nr_new_elements);
+    LIBSEMIGROUPS_ASSERT(_lenindex.size() > 1);
+    _lenindex[1] += nr_new_elements;
+    _left.add_cols(m);
+    _reduced.add_cols(m);
+    _right.add_cols(m);
+  }
 
+  TEMPLATE
+  template <typename T>
+  void FROIDURE_PIN::add_generators_after_start(T const& first, T const& last) {
+    detail::Timer timer;
+    size_t const  tid = THREAD_ID_MANAGER.tid(std::this_thread::get_id());
+
+    // get some parameters from the old semigroup
+    letter_type const old_nrgens  = nr_generators();
+    size_type const   old_nr      = _nr;
+    size_type         nr_old_left = _pos;
+
+    // erase the old index
+    _enumerate_order.erase(_enumerate_order.begin() + _lenindex[1],
+                           _enumerate_order.end());
+
+    add_generators_before_start(first, last);
+
+    // old_new[i] indicates if we have seen _elements.at(i) yet in new.
+    std::vector<bool> old_new(old_nr, false);
+
+    for (letter_type i = 0; i < _letter_to_pos.size(); i++) {
+      if (_letter_to_pos[i] < old_nr) {
+        old_new[_letter_to_pos[i]] = true;
+      } else {
+        break;
+      }
+    }
     // reset the data structure
     _idempotents_found = false;
     _nr_rules          = _duplicate_gens.size();
     _pos               = 0;
     _wordlen           = 0;
-    _nrgens            = _gens.size();
     _lenindex.clear();
     _lenindex.push_back(0);
-    _lenindex.push_back(_nrgens - _duplicate_gens.size());
-
-    // Add columns for new generators
-    _reduced = detail::DynamicArray2<bool>(
-        _nrgens, _reduced.nr_rows() + _nrgens - old_nrgens);
-    _left.add_cols(_nrgens - _left.nr_cols());
-    _right.add_cols(_nrgens - _right.nr_cols());
-
-    // Add rows in for newly added generators
-    _left.add_rows(_nrgens - old_nrgens);
-    _right.add_rows(_nrgens - old_nrgens);
+    _lenindex.push_back(nr_generators() - _duplicate_gens.size());
+    std::fill(_reduced.begin(), _reduced.end(), false);
 
     size_type nr_shorter_elements;
-
+    auto      ptr = _state.get();
     // Repeat until we have multiplied all of the elements of <old> up to the
     // old value of _pos by all of the (new and old) generators
 
@@ -1025,14 +685,14 @@ namespace libsemigroups {
               _nr_rules++;
             }
           }
-          for (letter_type j = old_nrgens; j < _nrgens; j++) {
-            closure_update(i, j, b, s, old_nr, tid, old_new);
+          for (letter_type j = old_nrgens; j < nr_generators(); j++) {
+            closure_update(i, j, b, s, old_nr, tid, old_new, ptr);
           }
         } else {
           // _elements[i] is either not in old, or it is in old but its
           // descendants are not known
-          for (letter_type j = 0; j < _nrgens; j++) {
-            closure_update(i, j, b, s, old_nr, tid, old_new);
+          for (letter_type j = 0; j < nr_generators(); j++) {
+            closure_update(i, j, b, s, old_nr, tid, old_new, ptr);
           }
         }
         _pos++;
@@ -1043,7 +703,7 @@ namespace libsemigroups {
         if (_wordlen == 0) {
           for (enumerate_index_type i = 0; i < _pos; i++) {
             size_t b = _final[_enumerate_order[i]];
-            for (letter_type j = 0; j < _nrgens; j++) {
+            for (letter_type j = 0; j < nr_generators(); j++) {
               // TODO(JDM) reuse old info here!
               _left.set(
                   _enumerate_order[i], j, _right.get(_letter_to_pos[j], b));
@@ -1053,7 +713,7 @@ namespace libsemigroups {
           for (enumerate_index_type i = _lenindex[_wordlen]; i < _pos; i++) {
             element_index_type p = _prefix[_enumerate_order[i]];
             letter_type        b = _final[_enumerate_order[i]];
-            for (letter_type j = 0; j < _nrgens; j++) {
+            for (letter_type j = 0; j < nr_generators(); j++) {
               // TODO(JDM) reuse old info here!
               _left.set(_enumerate_order[i], j, _right.get(_left.get(p, j), b));
             }
@@ -1073,6 +733,42 @@ namespace libsemigroups {
     report_why_we_stopped();
   }
 
+  VOID FROIDURE_PIN::add_generator(const_reference x) {
+    if (immutable()) {
+      LIBSEMIGROUPS_EXCEPTION("cannot add generators, the FroidurePin instance "
+                              "has been set to immutable");
+    }
+    validate_element(x);
+    if (_pos == 0) {
+      add_generators_before_start(&x, &x + 1);
+    } else {
+      add_generators_after_start(&x, &x + 1);
+    }
+  }
+
+  TEMPLATE
+  template <typename T>
+  void FROIDURE_PIN::add_generators(T const& first, T const& last) {
+    if (immutable()) {
+      LIBSEMIGROUPS_EXCEPTION("cannot add generators, the FroidurePin instance "
+                              "has been set to immutable");
+    }
+    validate_element_collection(first, last);
+    if (_pos == 0) {
+      add_generators_before_start(first, last);
+    } else {
+      add_generators_after_start(first, last);
+    }
+  }
+
+  TEMPLATE
+  template <typename T>
+  void FROIDURE_PIN::add_generators(T const& coll) {
+    static_assert(!std::is_pointer<T>::value,
+                  "the template parameter T must not be a pointer");
+    add_generators(coll.begin(), coll.end());
+  }
+
   VOID
   FROIDURE_PIN::add_generators(std::initializer_list<const_element_type> coll) {
     add_generators<std::initializer_list<const_element_type>>(coll);
@@ -1104,7 +800,7 @@ namespace libsemigroups {
     } else {
       for (const_reference x : coll) {
         if (!contains(x)) {
-          add_generators({x});
+          add_generator(x);
         }
       }
     }
@@ -1134,38 +830,6 @@ namespace libsemigroups {
     }
   }
 
-  BOOL FROIDURE_PIN::is_monoid() {
-    run();
-    return _found_one;
-  }
-
-  // More specialisations below in kbe.*pp and element.*pp
-  TRIL FROIDURE_PIN::is_finite() {
-    return tril::TRUE;
-  }
-
-  ////////////////////////////////////////////////////////////////////////
-  // FroidurePin - validation member functions - private
-  ////////////////////////////////////////////////////////////////////////
-
-  VOID FROIDURE_PIN::validate_element_index(element_index_type i) const {
-    if (i >= _nr) {
-      LIBSEMIGROUPS_EXCEPTION(
-          "element index out of bounds, expected value in [0, %d), got %d",
-          _nr,
-          i);
-    }
-  }
-
-  VOID FROIDURE_PIN::validate_letter_index(letter_type i) const {
-    if (i >= nr_generators()) {
-      LIBSEMIGROUPS_EXCEPTION(
-          "generator index out of bounds, expected value in [0, %d), got %d",
-          nr_generators(),
-          i);
-    }
-  }
-
   ////////////////////////////////////////////////////////////////////////
   // FroidurePin - enumeration member functions - private
   ////////////////////////////////////////////////////////////////////////
@@ -1191,12 +855,15 @@ namespace libsemigroups {
     }
   }
 
-  // _nrgens, _duplicates_gens, _letter_to_pos, and _elements must all be
-  // initialised for this to work, and _gens must point to an empty vector.
-  VOID FROIDURE_PIN::copy_gens() {
+  // _duplicates_gens, _letter_to_pos, and _elements must all be
+  // initialised for this to work, and _gens must be an empty vector.
+  VOID FROIDURE_PIN::copy_generators_from_elements(size_t N) {
     LIBSEMIGROUPS_ASSERT(_gens.empty());
-    _gens.resize(_nrgens);
-    std::vector<bool> seen(_nrgens, false);
+    if (N == 0) {
+      return;
+    }
+    _gens.resize(N);
+    std::vector<bool> seen(N, false);
     // really copy duplicate gens from _elements
     for (std::pair<letter_type, letter_type> const& x : _duplicate_gens) {
       // The degree of everything in _elements has already been increased (if
@@ -1206,7 +873,7 @@ namespace libsemigroups {
       seen[x.first]  = true;
     }
     // the non-duplicate gens are already in _elements, so don't really copy
-    for (letter_type i = 0; i < _nrgens; i++) {
+    for (letter_type i = 0; i < N; i++) {
       if (!seen[i]) {
         _gens[i] = _elements[_letter_to_pos[i]];
       }
@@ -1219,7 +886,8 @@ namespace libsemigroups {
                                     element_index_type s,
                                     size_type          old_nr,
                                     size_t const&      tid,
-                                    std::vector<bool>& old_new) {
+                                    std::vector<bool>& old_new,
+                                    state_type*        ptr) {
     if (_wordlen != 0 && !_reduced.get(s, j)) {
       element_index_type r = _right.get(s, j);
       if (_found_one && r == _pos_one) {
@@ -1230,10 +898,11 @@ namespace libsemigroups {
         _right.set(i, j, _right.get(_letter_to_pos[b], _final[r]));
       }
     } else {
-      Product()(this->to_external(_tmp_product),
-                this->to_external_const(_elements[i]),
-                this->to_external_const(_gens[j]),
-                tid);
+      InternalProduct()(this->to_external(_tmp_product),
+                        this->to_external_const(_elements[i]),
+                        this->to_external_const(_gens[j]),
+                        ptr,
+                        tid);
       auto it = _map.find(_tmp_product);
       if (it == _map.end()) {  // it's new!
         is_one(_tmp_product, _nr);
@@ -1478,14 +1147,16 @@ namespace libsemigroups {
     // Cannot use _tmp_product itself since there are multiple threads here!
     internal_element_type tmp_product = this->internal_copy(_tmp_product);
     size_t tid = THREAD_ID_MANAGER.tid(std::this_thread::get_id());
+    auto   ptr = _state.get();
 
     for (; pos < last; pos++) {
       element_index_type k = _enumerate_order[pos];
       if (_is_idempotent[k] == 0) {
-        Product()(this->to_external(tmp_product),
-                  this->to_external(_elements[k]),
-                  this->to_external(_elements[k]),
-                  tid);
+        InternalProduct()(this->to_external(tmp_product),
+                          this->to_external(_elements[k]),
+                          this->to_external(_elements[k]),
+                          ptr,
+                          tid);
         if (InternalEqualTo()(tmp_product, _elements[k])) {
           idempotents.emplace_back(_elements[k], k);
           _is_idempotent[k] = 1;
